
from typing import TypeVar
from logging import getLogger
import warnings

from extra_data.read_machinery import select_train_ids


log = getLogger('exdf.data_reduction.ReductionMethod')
train_sel = TypeVar('train_sel')
index_exp = TypeVar('index_exp')


def is_train_selection(x):
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            select_train_ids([], x)
    except TypeError:
        return False
    else:
        return True


def is_index_expression(x):
    print(x, type(x))
    if isinstance(x, (slice, list)):
        return True

    if isinstance(x, tuple) and all([isinstance(y, (slice, int)) for y in x]):
        return True

    return False


class ReductionMethod(list):
    """A method generating data reduction operations."""

    def __init__(self, data, ap):
        # Methods should always implement their own __init__
        raise NotImplementedError('__init__')

    def _emit(self, *args):
        self.append(args)

        cls = self.__class__
        log.debug(f'<{cls.__module__}.{cls.__qualname__}> emitted {args}')

    def select_trains(
        self,
        source_glob: str,
        trains: train_sel
    ):
        assert isinstance(source_glob, str)
        assert is_train_selection(trains)
        self._emit('select-trains', source_glob, trains)

    def select_pulses(
        self,
        source_glob: str,
        index_group: str,  # May be xtdf
        trains: train_sel,
        rows: index_exp
    ):
        raise NotImplementedError('select-pulses')
        assert isinstance(source_glob, str)
        assert isinstance(index_group, str)
        assert is_train_selection(trains)
        assert is_index_expression(rows)
        self._emit('select-pulses', source_glob, index_group, trains, rows)

    def remove_sources(
        self,
        source_glob: str
    ):
        assert isinstance(source_glob, str)
        self._emit('remove-sources', source_glob)

    def remove_keys(
        self,
        source_glob: str,
        key_glob: str
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        self._emit('remove-keys', source_glob, key_glob)

    def rechunk_keys(
        self,
        source_glob: str,
        key_glob: str,
        chunks: tuple
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert isinstance(chunks, tuple)
        assert all([x is None or isinstance(x, int) for x in chunks])
        self._emit('rechunk-keys', source_glob, key_glob, chunks)

    def partial_copy(
        self,
        source_glob: str,
        key_glob: str,
        region: index_exp
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert is_index_expression(region)
        self._emit('partial-copy', source_glob, key_glob, region)

    @staticmethod
    def arguments(ap):
        pass  # Reduction method may not require any arguments.
